"""
A library to generate Matplotilib graphics within django applications
"""

from .views import (
        PNGPlotView,
        SVGZPlotView,
        CachedPNGPlotView,
        CachedSVGZPlotView,
        PNGViewPlot,
        SVGViewPlot,
        )
from .value import (
        SVGPlotToValue,
        PNGBase64PlotToValue,
        CachedSVGPlotToValue,
        CachedPNGBase64PlotToValue,
        SVGValuePlot,
        PNGValuePlot,
        )
from .file import (
        SVGZPlotToFile,
        PNGPlotToFile,
        SVGFilePlot,
        PNGFilePlot,
        )

__all__ = ["PNGPlotView",
           "SVGZPlotView",
           "PNGPlotToFile",
           "SVGZPlotToFile",
           "SVGPlotToValue",
           "PNGBase64PlotToValue"
           "CachedPNGPlotView",
           "CachedSVGZPlotView",
           "CachedSVGPlotToValue",
           "PNGBase64PlotToValue"
           "PNGViewPlot",
           "SVGViewPlot",
           "PNGFilePlot",
           "SVGFilePlot",
           "PNGValuePlot",
           "SVGValuePlot",
           ]

__version__ = "0.0.3"



"""
    ========
    value.py
    ========

    This module provides with classes that assist in storing plots into
    variables ready to be passed to Django templates and include the plot
    within the html document.
"""
from base64 import b64encode
from django.utils.safestring import mark_safe
from .base import (
        BasePlot,
        CachedMixin,
        SVGPlotMixin,
        PNGPlotMixin,
)


class ValueMixin:
    """
    This mixins provides ``BasePlot`` with the ``get_value()`` method that
    returns a safe string to be used inside a template. It also has the magic
    method ``__str__.py`` so it can be inyected to the template.
    """

    def get_value(self):
        """
        Add the returned value of this method to the context dictionary that is
        passed to render the template.
        """
        image_buffer = self.get_image()
        value = image_buffer.getvalue()
        return mark_safe(value)

    def __str__(self):
        return self.get_value()


class Base64ValueMixin:
    """
    This mixin provides the ``BasePlot`` class with base64 encoding of its
    provided value through ``get_value()`` method. It also has the magic
    method ``__str__.py`` so it can be inyected to the template.

    """
    def get_value(self):
        """
        Add the returned value of this method to the context dictionary that is
        passed to render the template.
        """
        image_buffer = self.get_image()
        value = image_buffer.getvalue()
        return mark_safe(b64encode(value).decode("utf-8"))


class SVGPlotToValue(SVGPlotMixin, ValueMixin, BasePlot):
    """
    This class is used to create a SVG figure into a variable ready to
    be passed to the template engine to render it in a webpage.
    """
    pass


class PNGBase64PlotToValue(Base64ValueMixin, PNGPlotMixin, ValueMixin,
                           BasePlot):
    """
    This class is used to dump a PNG figure encoded in Base64 format into a
    variable ready to be passed to the template engine to render it in a
    webpage.
    """
    pass


class CachedSVGPlotToValue(CachedMixin, SVGPlotMixin, ValueMixin, BasePlot):
    """
    This class is used to create a SVG figure into a variable ready to
    be passed to the template engine to render it in a webpage. Cached Version.
    """
    pass


class CachedPNGBase64PlotToValue(CachedMixin, Base64ValueMixin, PNGPlotMixin,
                                 ValueMixin, BasePlot):
    """
    This class is used to dump a PNG figure encoded in Base64 format into a
    variable ready to be passed to the template engine to render it in a
    webpage. Cached Version.
    """
    pass


SVGValuePlot = CachedSVGPlotToValue
PNGValuePlot = CachedPNGBase64PlotToValue



"""
    =======
    base.py
    =======

    This module contains the basic classes used for building the actual classes
    that are used in generating the plots. It is structured around a base class
    called ``BasePlot`` that implements the common logic required for the
    project operations with some mixin_ classes to add it extra functionality
    by multiple inheritance.

    .. _mixin: https://en.wikipedia.org/wiki/Mixin
"""

from typing import Any
from io import BytesIO, StringIO
from contextlib import contextmanager
import matplotlib.pyplot as plt
from django.core.cache import caches


class BasePlot:
    """
    Base class of all the plot classes used in this library. It implements the
    basic protocol that gathers the data, draws the graph and passes it to an
    in memory file like object ready to be consumed.
    """
    buffer_class: Any = BytesIO
    file_format = ""

    @staticmethod
    def plotter_function(data, **kwargs):
        """
        Override this method with an **statically linked function** that
        returns a Matplotlib figure to be lately used to render the plot
        in given file formats.

        :data: A data structure that contains the information to be graphically
            modeled by the plotter function.

        :kwargs: A dict of parameters used to customize the image rendering.

        """
        raise NotImplementedError("Plot class requires a plotter method")

    def get_plot_data(self):
        """
        Override this method to provide data to ``plotter_function()``
        method.
        """
        return {}

    def get_plot_options(self):
        """
        Override this method to provide the plotter function with extra
        parameters to customize the graphical generation.
        """
        return {}

    def get_filetype(self):
        """
        Override this method to dinamically set the file format of the
        generated file.
        """
        return self.file_format

    @contextmanager
    def _get_figure(self):
        data = self.get_plot_data()
        plot_options = self.get_plot_options()
        figure = self.plotter_function(data, **plot_options)
        try:
            yield figure
        finally:
            plt.close(figure)

    def process_image(self, image_buffer):
        """
        Override this method to add modifications to the image such as
        watermarks before caching.
        """
        return image_buffer

    def get_image(self):
        """
        Returns a in memory file object with the plot image.
        """
        image_buffer = self.buffer_class()
        with self._get_figure() as figure:
            figure.savefig(image_buffer, format=self.get_filetype())
        image_buffer = self.process_image(image_buffer)
        image_buffer.seek(0)
        return image_buffer


class CachedMixin:
    """
    Mixin class to add cache functionality to the BasePlot object.
    """
    cache_backend_name = "default"
    cache_timeout = -1

    def get_cache_key(self):
        """
        Override this method with a value that changes when plot regeneration
        is required.
        """
        raise NotImplementedError

    def get_image(self):
        cache_backend = caches[self.cache_backend_name]
        cache_key = self.get_cache_key()
        image_value = cache_backend.get(cache_key)

        if image_value is None:
            image_file = super().get_image()
            image_value = image_file.getvalue()
            if self.cache_timeout == -1:
                cache_backend.set(cache_key, image_value)
            else:
                cache_backend.set(cache_key, image_value, self.cache_timeout)
            return image_file
        else:
            return self.buffer_class(image_value)


class SVGPlotMixin(BasePlot):
    """
    Mixing to generate a memory file containing an SVG figure.
    """
    buffer_class = StringIO
    file_format = "svg"


class SVGZPlotMixin(BasePlot):
    """
    Mixin to generate a memory file containing an SVGZ figure.
    """
    buffer_class = BytesIO
    file_format = "svgz"


class PNGPlotMixin(BasePlot):
    """
    Mixin to generate a memory file containing a binary PNG figure.
    """
    buffer_class = BytesIO
    file_format = "png"



"""
    =======
    file.py
    =======

    This module provides the classes required to generate Django files ready to
    be stored in media directories.
"""

from django.core.files import File as DjangoFile
from .base import BasePlot, SVGZPlotMixin, PNGPlotMixin


class FileMixin:
    """
    This mixin provides the ``BasePlot`` class with the capability of
    generating a in memory file to be passed to an Image field.
    """
    file_class = DjangoFile
    filename = ""

    def get_filename(self):
        """
        Override this method to dinamically change the name of the file object
        created by the ``get_file()`` method.
        """
        name = self.filename
        ext = self.get_filetype()
        return f"{name}.{ext}"

    def get_file(self):
        """
        This method returns a Django file object ready to be assigned to an
        ImageField.
        """
        name = self.get_filename()
        file = self.file_class(self.get_image(), name)
        return file


class SVGZPlotToFile(SVGZPlotMixin, FileMixin, BasePlot):
    """
    This class is used to create a Django a ``File`` object storing an SVGZ
    plot.
    """
    pass


class PNGPlotToFile(PNGPlotMixin, FileMixin, BasePlot):
    """
    This class is used to create a Django a ``File`` object storing a PNG plot.
    """
    pass


SVGFilePlot = SVGZPlotToFile
PNGFilePlot = PNGPlotToFile



"""
    ========
    views.py
    ========

    This module provides with a collection of Django View classes that respond
    to requests with HttpResponse objects with Matplotlib figures as payload.

"""
from typing import Any
from django.views import View
from django.http import HttpResponse
from .base import CachedMixin, SVGZPlotMixin, PNGPlotMixin


class BasePlotView(View):
    """
    A Django View class that has the common logic of all project's View
    classes. It is not intended to be used in production but instead to be
    subclassed and extended via mixins.
    """
    http_response_class = HttpResponse
    buffer_class: Any = None
    file_format = ""
    filename = ""
    encoding = ""
    disposition = ""  # values: inline, attachment
    mimetype = ""
    http_method_names = [
            "get",
            "head",
            "options",
    ]

    def get_disposition(self):
        """
        Override this method if you want to dinamically set the content
        disposition of the ``HttpResponse`` object.
        """
        return self.disposition

    def get_filename(self):
        """
        Override this method if you want to dinamically set the name of
        the file generated from the figure when the disposition is attachment.
        """
        return self.filename

    def get_mimetype(self):
        """
        Override this method if you want to dinamically set the mimetype of
        the generated file.
        """
        return self.mimetype

    def get_encoding(self):
        """
        Override this method if you want to dinamically set the encoding of
        the generated file.
        """
        return self.encoding

    def get_headers(self, buffer):
        """
        Returns a dict of parameters to be used as headers of the response
        object. Override it to provide extend the values it contains.
        """
        result = {}

        disposition = self.get_disposition()
        if disposition:
            if disposition == "inline":
                result['Content-Disposition'] = 'inline'
            else:
                name = self.get_filename()
                value = f'attachment; filename="{name}"'
                result['Content-Disposition'] = value

        encoding = self.get_encoding()
        if encoding:
            result['Content-Encoding'] = encoding
        result['Content-Length'] = buffer.getbuffer().nbytes
        return result

    def get(self, request, *args, **kwargs):
        """
        This methods generates the GET response.
        """
        buffer = self.get_image()
        headers = self.get_headers(buffer)

        response = self.http_response_class(buffer,
                                            content_type=self.get_mimetype())
        for key, value in headers.items():
            response.headers[key] = value
        return response

    def head(self, request, *args, **kwargs):
        """
        This methods generates the HEAD response.
        """
        buffer = self.get_image()
        headers = self.get_headers(buffer)

        response = self.http_response_class(b"",
                                            content_type=self.get_mimetype())
        for key, value in headers.items():
            response.headers[key] = value
        return response


class PNGPlotView(PNGPlotMixin, BasePlotView):
    """
    A Django ``View`` class that returns a PNG graphic file as ``HttpResponse``
    payload.
    """
    disposition = "inline"
    mimetype = "image/png"


class SVGZPlotView(SVGZPlotMixin, BasePlotView):
    """
    A Django `View` class that returns a SVGZ graphic file as `HttpResponse`
    payload.
    """
    disposition = "inline"
    encoding = "gzip"
    mimetype = "image/svg+xml"


class CachedPNGPlotView(CachedMixin, PNGPlotMixin, BasePlotView):
    """
    A Django ``View`` class that returns a PNG graphic file as ``HttpResponse``
    payload.
    """
    disposition = "inline"
    mimetype = "image/png"


class CachedSVGZPlotView(CachedMixin, SVGZPlotMixin, BasePlotView):
    """
    A Django `View` class that returns a SVGZ graphic file as `HttpResponse`
    payload.
    """
    disposition = "inline"
    encoding = "gzip"
    mimetype = "image/svg+xml"


SVGViewPlot = CachedSVGZPlotView
PNGViewPlot = CachedPNGPlotView
